--[[
--
-- TODO
-- * Add a LDB that shows the assignments of everyone
-- * Add a countdown to when the next person in line should get close
-- * Support 10man (just use one list)
--
--]]

local f = CreateFrame("Frame")
local whisperNext = "TMP: You are getting the Spark next after %s!"
local whisperNow = "TMP: GET IT NOW!"
local warningNotScanned = "%s could not be scanned, you need to scan again with /tmpscan when that player is closer."
local warningMoreToScan = "There are still players left to be scanned!"
local warningScanDone = "TMP scan has completed."

local tormentedIds = {
	[99489] = true,
	[99257] = true,
	[99402] = true,
	[99403] = true,
	[99404] = true,
}

local tormentIds = {
	[99256] = true,
	[100230] = true,
	[100231] = true,
	[100232] = true,
}

local stacks = 9

local meleeAssignment = 1
local rangedAssignment = 2

local needsScanning = {}
local talentSpecs = {}

local function checkWarrior(player)
	if not talentSpecs[player] then return end
	if talentSpecs[player] == 3 then
		return
	end
	return 1
end

local function checkPaladin(player)
	if not talentSpecs[player] then return end
	if talentSpecs[player] == 2 or talentSpecs[player] == 1 then
		return
	end
	return 1
end

local function checkDeathknight(player)
	if not talentSpecs[player] then return end
	if talentSpecs[player] == 1 then
		return
	end
	return 1
end

local function checkDruid(player)
	if not talentSpecs[player] then return end
	if talentSpecs[player] == 2 then -- XXX Feral, check for buff or armor value?
		-- XXX assume ferals are tanks for now
		return
	elseif talentSpecs[player] == 3 then
		return
	elseif talentSpecs[player] == 1 then
		return 2
	end
end

local defaultAssignments = {
	MAGE = 2,
	WARLOCK = 2,
	HUNTER = 2,
	ROGUE = 1,
	WARRIOR = checkWarrior,
	PALADIN = checkPaladin,
	DEATHKNIGHT = checkDeathknight,
	DRUID = checkDruid,
}
local assignments = {}

local total = 0

local roomHeight, roomWidth = 1587.49993896484, 1058.3332824707
function getUnitPosition(unit)
	local x,y = GetPlayerMapPosition(unit)
	if x <= 0 and y <= 0 then
		SetMapToCurrentZone()
		x,y = GetPlayerMapPosition(unit)
	end
	return x,y
end
function getDistance(unit1,unit2)
	local x1,y1 = getUnitPosition(unit1)
	local x2,y2 = getUnitPosition(unit2)
	local dx = (x2 - x1) * roomWidth
	local dy = (y2 - y1) * roomHeight
	return (dx*dx + dy*dy)^(0.5) -- dx*dx is faster than dx^2
end

local tormented = GetSpellInfo(99403)
local torment = GetSpellInfo(99256)
local tormentedPlayers = {}
local nextTargets = {}

local tmpDistances = {}
local possibleTargets = {}
local function distanceSort(a, b) return (tmpDistances[a] or 0) < (tmpDistances[b] or 0) end
local function getNextTarget(current, assignment)
	if not current or not assignment then return end
	wipe(tmpDistances)
	wipe(possibleTargets)
	for player, a in pairs(assignments) do
		if assignment == a and current ~= player and UnitInRaid(player) and not UnitIsDeadOrGhost(player) and not tormentedPlayers[player] then
			tmpDistances[player] = getDistance(current, player)
			possibleTargets[#possibleTargets + 1] = player
		end
	end
	table.sort(possibleTargets, distanceSort)
	if #possibleTargets == 0 then
		print("No possible targets, the stack needs to go higher.")
		return
	end
	return possibleTargets[1]
end

local function populateNextTarget(player)
	local nextTarget = getNextTarget(player, assignments[player])
	if nextTarget then
		nextTargets[player] = nextTarget
		SendChatMessage(whisperNext:format(player), "WHISPER", nil, nextTarget)
	else
		print("Could not find a replacement for " .. tostring(player) .. ".")
	end
end

function f:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, _, _, _, _, _, _, player, _, _, spellId, _, _, _, buffStack)
	if event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_APPLIED_DOSE" then
		if tormentedIds[spellId] then
			tormentedPlayers[player] = true
			-- XXX whisper player to GTFO on heroic?
		elseif tormentIds[spellId] then
			if buffStack == 1 or not nextTargets[player] then
				populateNextTarget(player)
			elseif buffStack > 7 then
				SendChatMessage(whisperNow, "WHISPER", nil, nextTargets[player])
				nextTargets[player] = nil
			end
		end
	elseif event == "SPELL_AURA_REMOVED" and tormentedIds[spellId] then
		tormentedPlayers[player] = nil
	end
end

do
	local lastUnitQueried = nil
	local lastScanTime = 0
	local total = 0

	hooksecurefunc("NotifyInspect", function(unit)
		lastUnitQueried = UnitName(unit)
		lastScanTime = GetTime()
	end)
	local function scanNext()
		if lastUnitQueried then return end
		local unit = next(needsScanning)
		if not unit then
			print(warningScanDone)
			return
		end
		if UnitExists(unit) and UnitIsVisible(unit) and UnitIsConnected(unit) and CheckInteractDistance(unit, 4) then
			if CanInspect(unit) and GetTime() > (lastScanTime + 3) then
				needsScanning[unit] = nil
				NotifyInspect(unit)
			end
		end
	end

	f:SetScript("OnUpdate", function(self, elapsed)
		total = total + elapsed
		if total >= 4 then
			total = 0
			scanNext()
		end
	end)
	f:Hide()

	local function getUnitNameFromGUID(guid)
		if UnitInRaid("player") then
			for i = 1, GetNumRaidMembers() do
				local n = GetRaidRosterInfo(i)
				if n and UnitGUID(n) == guid then return n end
			end
		else
			for i = 1, GetNumPartyMembers() do
				local pId = UnitGUID("party" .. i)
				if pId == guid then return UnitName("party" .. i) end
			end
		end
	end

	function f:INSPECT_READY(event, guid)
		if not lastUnitQueried or guid ~= UnitGUID(lastUnitQueried) then
			lastUnitQueried = getUnitNameFromGUID(guid)
		end
		if not lastUnitQueried then return end
		local group = GetActiveTalentGroup(true)
		local tree = 1
		for i = 1, 3 do
			local _,_,_,_, pointsSpent = GetTalentTabInfo(i, true, nil, group)
			if pointsSpent > 30 then
				tree = i + 1
				break
			end
		end
		local _, class = UnitClass(lastUnitQueried)
		if class then
			talentSpecs[lastUnitQueried] = tree
			local def = defaultAssignments[class]
			assignments[lastUnitQueried] = type(def) == "function" and def(lastUnitQueried) or def
			needsScanning[lastUnitQueried] = nil
		end
		lastUnitQueried = nil
	end

	function f:RAID_ROSTER_UPDATE()
		for i = 1, GetNumRaidMembers() do
			local n,_,_,_,_,class = GetRaidRosterInfo(i)
			if n and not assignments[n] then
				local def = defaultAssignments[class]
				if type(def) == "function" then
					if talentSpecs[n] then
						assignments[n] = def(n)
					elseif lastUnitQueried ~= n then
						needsScanning[n] = true
					end
				else
					assignments[n] = def
				end
			end
		end
		for u in pairs(assignments) do
			if not UnitExists(u) then
				assignments[u] = nil
				talentSpecs[u] = nil
				needsScanning[u] = nil
			end
		end
	end
end

do
	local function checkBaleroc(unit)
		local guid = UnitGUID(unit)
		if not guid then return end
		local id = tonumber(guid:sub(7, 10), 16)
		if id ~= 53494 then return end -- We're not engaging Baleroc
		if next(needsScanning) then
			print(warningMoreToScan)
		end
	end
	function f:INSTANCE_ENCOUNTER_ENGAGE_UNIT() checkBaleroc("boss1") end
	function f:PLAYER_TARGET_CHANGED() checkBaleroc("target") end
end

f:RegisterEvent("PLAYER_TARGET_CHANGED")
f:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
f:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
f:RegisterEvent("INSPECT_READY")
f:RegisterEvent("RAID_ROSTER_UPDATE")
f:SetScript("OnEvent", function(self, event, ...) self[event](self, event, ...) end)

local function filter(self, event, msg)
	if msg:find("^TMP: ") then return true end
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", filter)

SlashCmdList.TormentMePlz = function() f:Show() end
SLASH_TormentMePlz1 = "/tmpscan"
